// jointAnchor.cpp  by Kosei Demura 2007-11-11
// My web site is http://demura.net
// This program uses the Open Dynamics Engine (ODE) by Russell Smith.
// The ODE web site is http://ode.org/

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

#ifdef dDOUBLE
#define dsDrawBox      dsDrawBoxD
#define dsDrawSphere   dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCapsule  dsDrawCapsuleD
#define dsDrawLine     dsDrawLineD
#endif

#define BODY_NUM       4   //  the number of body
#define BODY_LENGTH  0.5   // body legnth
#define BODY_RADIUS  0.05   // body radius
#define BODY_WEIGHT  1.0   // body weight


static dWorldID world;
static dSpaceID space;  // add
static dGeomID  ground; // add 
static dJointGroupID contactgroup; // add
dsFunctions fn;

typedef struct {    
  dBodyID body;     
  dGeomID geom;     
  dReal   radius;
  dReal   length;
  dReal   mass;
} MyLink;
MyLink leg[BODY_NUM];      // add

dJointID joint[BODY_NUM-1];

int drawBody = 1;
dReal angle  = 0.0;
static int   steps = 0;

static void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
  static const int N = 7;
  dContact contact[N];
  
  // exit without doing anything if the two bodies are connected by a joint
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnected (b1,b2))  return;

  int isGround = ((ground == o1) || (ground == o2));

  int n =  dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
  // printf("n=%d \n",n);
  if (isGround)  {
    for (int i = 0; i < n; i++) {
			contact[i].surface.mode = dContactBounce | dContactMu2;
			contact[i].surface.mu    = 1.0;
			contact[i].surface.mu2   = 0.1;				
 
      contact[i].surface.bounce     = 0.1; // (0.0~1.0)
			contact[i].surface.bounce_vel = 0.1; 
      dJointID c = dJointCreateContact(world,contactgroup,&contact[i]);
      dJointAttach (c,dGeomGetBody(contact[i].geom.g1),
		    dGeomGetBody(contact[i].geom.g2));
    }
  }
}

void control(int i, dJointID ajoint)
{
 	static dReal ang = 0.0; 
  dReal velCoeff   = 10.0;
  dReal force      = 100.0;
  dReal p,z;

	if (i % 2) angle =   0.5 * M_PI * sin(steps * M_PI/ 180.0);
	else       angle = - 0.5 * M_PI * sin(steps * M_PI/ 180.0 + 0.25 * M_PI);
  p = dJointGetHingeAngle(ajoint);
  z = angle - p;
  dJointSetHingeParam(ajoint, dParamVel, velCoeff*z);
  dJointSetHingeParam(ajoint, dParamFMax, force);
}


static void simLoop (int pause)
{

	const dReal *pos[BODY_NUM],*R[BODY_NUM];
	dMatrix3 R1;
	dVector3 ap[BODY_NUM-1];          // anchor point
	float    ar = 0.06;               // anchor radius

  if (!pause) {
   	if (steps++ < 10) dBodyAddForce (leg[0].body, 0, 1.0, 0);
		for (int i = 0; i < BODY_NUM-1; i++) control(i, joint[i]);
  	dSpaceCollide(space,0,&nearCallback);  // add, Write this first
  	dWorldStep(world,0.01);
  	dJointGroupEmpty(contactgroup); // add
	}

  // calculate positions and orientaitons
 	for (int i = 0; i < BODY_NUM; i++) {		
  	pos[i] = dBodyGetPosition(leg[i].body);
  	R[i]   = dBodyGetRotation(leg[i].body);
  }
	
	// calculate anchor points
	dRSetIdentity(R1);
	for (int i = 0; i < BODY_NUM-1; i++) {
		dJointGetHingeAnchor(joint[i], ap[i]);
	}
	
  // draw lines
 	for (int i = 0; i < BODY_NUM-1; i++) {
		dsSetColorAlpha(1.3, 1.3, 1.3, 1.0);
		dsDrawLine(pos[i], ap[i]);
		dsSetColorAlpha(0.0, 0.0, 0.0, 1.0);
		dsDrawLine(ap[i],  pos[i+1]);
	}

	dsSetColorAlpha(1.3, 0.0, 0.0, 0.5);
	// draw anchors
	for (int i = 0; i < BODY_NUM-1; i++) {
		dJointGetHingeAnchor(joint[i], ap[i]);
		dsDrawSphere(ap[i], R1, ar);	
	}
	
	// draw legs
	dsSetColorAlpha(0.0, 0.0, 1.3, 0.5);
	for (int i = 0; i < BODY_NUM; i++) {
	 	if (drawBody) dsDrawCapsule(pos[i],R[i],leg[i].length,leg[i].radius);
  }
}

void start()
{
  static float xyz[3] = {   3.5,0.0,1.0};
  static float hpr[3] = {-180.0,0.0,0.0};
  dsSetViewpoint(xyz,hpr);
}

// Create an object 
void createMonoBot() {
  dMass m1;
  dReal x0 = 0.0, y0 = 0.0;
	dReal z0 = 2.0, tmpz = 0.0;
	
  // leg
	tmpz = z0;
	for (int i = 0; i < BODY_NUM; i++) {
		leg[i].radius = BODY_RADIUS;
		leg[i].length = BODY_LENGTH;
   	leg[i].mass   = BODY_WEIGHT;
  	leg[i].body   = dBodyCreate(world);   
  	dMassSetZero(&m1);
  	dMassSetCapsuleTotal(&m1,leg[i].mass,3,leg[i].radius,leg[i].length);
  	dBodySetMass(leg[i].body,&m1);    		     
  	dBodySetPosition(leg[i].body, x0, y0, tmpz); 
    tmpz += leg[i].length + 2 * leg[i].radius;   
  	leg[i].geom = dCreateCapsule(space,leg[i].radius,leg[i].length); 
  	dGeomSetBody(leg[i].geom,leg[i].body);   
	}

	tmpz = z0 + 0.5 * (leg[0].length + leg[0].radius);
 	for (int i = 0; i < BODY_NUM-1; i++) {
  	// hinge joint
  	joint[i] = dJointCreateHinge(world, 0);	
  	dJointAttach(joint[i], leg[i].body,leg[i+1].body);
  	dJointSetHingeAnchor(joint[i], x0, y0, tmpz);
		tmpz += leg[i+1].length + 2 * leg[i+1].radius;
  	dJointSetHingeAxis(joint[i], 0, 1, 0);
  	dJointSetHingeParam(joint[i], dParamLoStop, -0.5 * M_PI);
  	dJointSetHingeParam(joint[i], dParamHiStop,  0.5 * M_PI);
	}
}

void command(int cmd)
{
	dReal f = 50;
  switch (cmd) {
    case 'f': angle += 0.1; break;
    case 'j': angle -= 0.1; break;
    case 'b': 
			  if (drawBody) drawBody = 0;
				else          drawBody = 1;
				break;
  }
}


void  setDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = &command;
  fn.stop    = NULL;
  fn.path_to_textures = "../../drawstuff/textures";
}

int main (int argc, char **argv)
{
  setDrawStuff();

  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0); 

  dWorldSetGravity(world,0,0,-9.8);
	dWorldSetERP(world, 0.9);
	dWorldSetCFM(world, 1e-5);

  // Create a ground
  ground = dCreatePlane(space,0,0,1,0);

  // create an object
  createMonoBot();

  dsSimulationLoop (argc,argv,640,480,&fn);

  dWorldDestroy (world);

  return 0;
}
